import cv2
import numpy as np
from PIL import Image
import os
from matplotlib import pyplot as plt



def align_image_horizontally(image: Image.Image, threshold: int = 50, minLineLength: int = 1600, maxLineGap: int = 122, show_lines: int = 0) -> Image.Image:
    print(f"Threshold: {threshold}, MinLineLength: {minLineLength}, MaxLineGap: {maxLineGap}, ShowLines: {show_lines}")
    # threshold: int = 50, minLineLength: int = 150, maxLineGap: int = 10
    # Преобразуем изображение в массив NumPy
    image_np = np.array(image)

    # Убедимся, что изображение в градациях серого (если изначально цветное)
    if len(image_np.shape) == 3:
        image_np = cv2.cvtColor(image_np, cv2.COLOR_BGR2GRAY)

    # Повышение контрастности изображения для лучшего выделения линий
    #image_np = cv2.equalizeHist(image_np)

    # Применяем медианное размытие для удаления мелких шумов и повышения четкости контуров
    smoothed_image = cv2.medianBlur(image_np, 5)

    # Применяем адаптивное пороговое преобразование для более точного выделения контуров
    binary_image = cv2.adaptiveThreshold(smoothed_image, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, 
                                        cv2.THRESH_BINARY_INV, blockSize=11, C=2)

    # Применяем морфологическое расширение для соединения близких элементов
    kernel = np.ones((3, 3), np.uint8)
    binary_image = cv2.dilate(binary_image, kernel, iterations=2)

    # Применяем морфологическое закрытие для объединения разорванных линий и удаления мелких пустот
    kernel = np.ones((5, 5), np.uint8)
    binary_image = cv2.morphologyEx(binary_image, cv2.MORPH_CLOSE, kernel)

    # Применяем размытие по Гауссу для дополнительного сглаживания перед детектированием краев
    binary_image = cv2.GaussianBlur(binary_image, (3, 3), 0)


    # Используем метод Canny для детектирования краев
    edges = cv2.Canny(binary_image, 150, 200, apertureSize=3)

    # Используем преобразование Хафа для поиска линий с расширенными параметрами
    lines = cv2.HoughLinesP(edges, rho=2, theta=np.pi / 3960, threshold=threshold, minLineLength=minLineLength, maxLineGap=maxLineGap)
  
    # Фильтрация линий по углу
    filtered_lines = []
    if lines is not None:
        for line in lines:
            x1, y1, x2, y2 = line[0]
            angle = round(np.arctan2(y2 - y1, x2 - x1) * 180 / np.pi, 2)
            if abs(angle) <= 2 or abs(angle - 180) <= 2:
                filtered_lines.append(line)

    
    if show_lines == 1:
        # Копируем изображение для отрисовки линий
        output_image = np.copy(image)
        if filtered_lines:
            for line in filtered_lines:
                x1, y1, x2, y2 = line[0]
                cv2.line(output_image, (x1, y1), (x2, y2), (111,), 7)
        # Показываем результат
        plt.subplot, plt.imshow(output_image, cmap='gray'), plt.title('Lines Detected')
        plt.show()
        if filtered_lines:
            print(f"Find {len(filtered_lines)} линий.")
        else:
            print("Линии не найдены.")
        
        
    
        
    
    
    if filtered_lines:
        angles = []
        iterations = 1  # Количество итераций

        for _ in range(iterations):
            if filtered_lines:
                angles = []
                for line in filtered_lines:
                    x1, y1, x2, y2 = line[0]
                    angle = round(np.arctan2(y2 - y1, x2 - x1) * 180 / np.pi, 2)

                    # Фильтруем только линии близкие к горизонтали (угол близкий к 0)
                    if abs(angle) <= 2 or abs(angle - 180) <= 2:
                        angles.append(angle)

                if angles:
                    # Рассчитываем средний угол
                    average_angle = np.mean(angles)

                    # Поворачиваем изображение на вычисленный угол
                    (h, w) = image_np.shape[:2]
                    center = (w // 2, h // 2)
                    rotation_matrix = cv2.getRotationMatrix2D(center, average_angle, 1.0)  # Вращение на вычисленный угол
                    # Вращаем изображение с белым фоном
                    aligned_image = cv2.warpAffine(image_np, rotation_matrix, (w, h), flags=cv2.INTER_CUBIC, 
                                                   borderMode=cv2.BORDER_CONSTANT, borderValue=255)

                    # Обновляем изображение
                    image_np = aligned_image

                    
                else:
                    print("No horizontal lines detected.")
                    break
            else:
                print("No lines detected.")
                break

        # Преобразуем обратно в изображение PIL
        return Image.fromarray(image_np)
    else:
        print("No lines detected.")
        return image
    
def find_large_clusters(image: np.ndarray, min_area: int, max_area: int, height: int = 300, width: int = 300, show_clusters: int = 0) -> list:	

    """Найти большие кластеры на изображении с объединением пересекающихся и включенных кластеров"""
    
    # Преобразуем изображение в бинарное
    _, binary_image = cv2.threshold(image, 128, 255, cv2.THRESH_BINARY_INV)
    
    # Выполняем дилатацию для соединения близких объектов
    kernel = np.ones((height, width), np.uint8)
    dilated_image = cv2.dilate(binary_image, kernel, iterations=1)
    
    # Находим контуры объектов
    contours, _ = cv2.findContours(dilated_image, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    clusters = []
    
    for contour in contours:
        area = cv2.contourArea(contour)
        if min_area <= area <= max_area:
            x, y, w, h = cv2.boundingRect(contour)
            clusters.append((x, y, x + w, y + h))

    # Функции для обработки кластеров
    def rects_intersect(r1, r2):
        return not (r1[2] < r2[0] or r1[0] > r2[2] or r1[3] < r2[1] or r1[1] > r2[3])

    def rect_includes(r1, r2):
        return r1[0] <= r2[0] and r1[1] <= r2[1] and r1[2] >= r2[2] and r1[3] >= r2[3]

    def merge_rects(r1, r2):
        return (min(r1[0], r2[0]), min(r1[1], r2[1]), max(r1[2], r2[2]), max(r1[3], r2[3]))

    # Объединение пересекающихся кластеров
    def handle_intersections(clusters):
        new_clusters = []
        while clusters:
            current = clusters.pop(0)
            merged = False
            for i in range(len(new_clusters)):
                if rects_intersect(current, new_clusters[i]):
                    new_clusters[i] = merge_rects(current, new_clusters[i])
                    merged = True
                    break
            if not merged:
                new_clusters.append(current)
        return new_clusters

    # Объединение включенных кластеров
    def handle_inclusions(clusters):
        new_clusters = []
        while clusters:
            current = clusters.pop(0)
            included = False
            for i in range(len(new_clusters)):
                if rect_includes(current, new_clusters[i]):
                    new_clusters[i] = current
                    included = True
                    break
                elif rect_includes(new_clusters[i], current):
                    included = True
                    break
            if not included:
                new_clusters.append(current)
        return new_clusters

    # Сначала обрабатываем пересечения, затем включения
    clusters = handle_intersections(clusters)
    clusters = handle_inclusions(clusters)

    if show_clusters == 1:
        # Визуализация
        result_image = binary_image.copy()  # Используем бинарное изображение
        for (x_min, y_min, x_max, y_max) in clusters:
            cv2.rectangle(result_image, (x_min, y_min), (x_max, y_max), (255), 22)  # Белый прямоугольник

        plt.imshow(result_image, cmap='gray')
        plt.title('Clusters on Image')
        plt.axis('off')
        plt.show()
        raise SystemExit

    return clusters
          
def create_image_with_clusters(original_image: np.ndarray, clusters: list) -> np.ndarray:
    """Создать изображение, объединяющее все выровненные кластеры на белом фоне"""
    # Создаем пустое изображение с белым фоном
    h, w = original_image.shape
    aligned_image = np.ones((h, w), dtype=np.uint8) * 255  # Белый фон

    for (x_min, y_min, x_max, y_max) in clusters:
        # Вырезаем кластер
        cluster_image = original_image[y_min:y_max, x_min:x_max]
        
        # Преобразуем кластер в PIL Image и выравниваем
        cluster_image_pil = Image.fromarray(cluster_image)
        aligned_cluster_image_pil = align_image_horizontally(cluster_image_pil)
        aligned_cluster_image = np.array(aligned_cluster_image_pil)
        
        # Помещаем выровненный кластер обратно на изображение
        aligned_image[y_min:y_max, x_min:x_max] = aligned_cluster_image
    
    return aligned_image

def save_image(image_np: np.ndarray, filename: str, dpi=(600, 600)):
    """Сохранить изображение в файл с указанным DPI"""
    if image_np.size == 0:
        raise ValueError("Image is empty")
    
    # Преобразуем numpy массив в PIL Image
    image_pil = Image.fromarray(image_np)
    
    # Сохраняем изображение с заданным DPI
    image_pil.save(filename, dpi=dpi)



def process_folder(input_folder: str, output_folder: str):
    """Обработать все изображения в указанной папке и сохранить их в другой папке"""
    # Создаем выходную папку, если она не существует
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)
    
    # Перебираем все файлы .bmp в входной папке
    for file_name in os.listdir(input_folder):
        if file_name.lower().endswith('.bmp'):
            input_path = os.path.join(input_folder, file_name)
            output_path = os.path.join(output_folder, file_name)
            
            # Открываем изображение
            image = Image.open(input_path)
            image_np = np.array(image)

            # Если изображение цветное, преобразуем в оттенки серого
            if len(image_np.shape) == 3:
                image_np = cv2.cvtColor(image_np, cv2.COLOR_RGB2GRAY)

            # Параметры для align_image_horizontally
            #threshold_value = 50
            #minLineLength_value = 1200
            #maxLineGap_value = 60
            #show_lines_value = 0

            # Параметры для find_large_clusters
            min_area = 10000
            max_area = 30000000
            height_value = 30
            width_value = 200
            show_clusters_value = 0

            # Вызов функций с передачей параметров

            # Найти большие кластеры
            clusters = find_large_clusters(image_np, min_area, max_area, height_value, width_value, show_clusters_value)

            # Выровнять изображение по горизонтали
            #aligned_image = align_image_horizontally(image, threshold_value, minLineLength_value, maxLineGap_value, show_lines_value)
           
            # Создать изображение с объединенными выровненными кластерами
            result_image = create_image_with_clusters(image_np, clusters)
            
   
            # Сохранить результат
            save_image(result_image, output_path, dpi=(600, 600))




# Пример вызова функции для обработки всех изображений в папке
input_folder = r'C:\Users\1\Desktop\Del'
output_folder = r'C:\Users\1\Desktop\Del\1'
process_folder(input_folder, output_folder)
